@@ -21,18 +21,6 @@ def parse_input_layer(keras_layer, input_names, input_shapes, data_reader, confi
     return layer, output_shape
 
 
-@keras_handler('Reshape')
-def parse_reshape_layer(keras_layer, input_names, input_shapes, data_reader, config):
-    assert(keras_layer["class_name"] == 'Reshape')
-
-    layer = parse_default_keras_layer(keras_layer, input_names)
-    
-    layer['target_shape'] = keras_layer['config']['target_shape']
-    output_shape = input_shapes[0][:1] + keras_layer['config']['target_shape']
-    
-    return layer, output_shape
-
-
 class BinaryQuantizer(Quantizer):
     def __init__(self, bits=2):
         if bits == 1:
@@ -82,7 +70,8 @@ def parse_dense_layer(keras_layer, input_names, input_shapes, data_reader, confi
     else:
         layer['weight_quantizer'] = None
         layer['bias_quantizer'] = None
-    output_shape = [input_shapes[0][0], layer['n_out']]
+    output_shape = input_shapes[0][:]
+    output_shape[-1] = layer['n_out']
 
     return layer, output_shape
 
@@ -97,23 +86,25 @@ def parse_activation_layer(keras_layer, input_names, input_shapes, data_reader,
     if layer['class_name'] != 'Activation':
         layer['activation'] = layer['class_name']
     if layer['class_name'] == 'LeakyReLU':
-        layer['activ_param'] = keras_layer["config"].get('alpha', 0.3)
+        layer['activ_param'] = keras_layer['config'].get('alpha', 0.3)
     elif layer['class_name'] == 'ThresholdedReLU':
-        layer['activ_param'] = keras_layer["config"].get('theta', 1.)
+        layer['activ_param'] = keras_layer['config'].get('theta', 1.)
     elif layer['class_name'] == 'ELU':
-        layer['activ_param'] = keras_layer["config"].get('alpha', 1.)
+        layer['activ_param'] = keras_layer['config'].get('alpha', 1.)
+    elif layer['class_name'] == 'ReLU':
+        layer['class_name'] = 'Activation'
 
     if layer['class_name'] == 'Activation' and layer['activation'] == 'softmax':
         layer['class_name'] = 'Softmax'
-    if layer['class_name'] == 'ReLU':
-        layer['class_name'] = 'Activation'
+    if layer['class_name'] == 'Softmax':
+        layer['axis'] = keras_layer['config'].get('axis', -1)
     
     return layer, [shape for shape in input_shapes[0]]
 
 
 @keras_handler('BatchNormalization')
 def parse_batchnorm_layer(keras_layer, input_names, input_shapes, data_reader, config):
-    assert('BatchNormalization' in keras_layer['class_name'])
+    assert('BatchNormalization' in keras_layer['class_name'] or 'QConv2DBatchnorm' in keras_layer['class_name'])
 
     layer = parse_default_keras_layer(keras_layer, input_names)
 
