@@ -112,6 +112,8 @@ def parse_default_keras_layer(keras_layer, input_names):
     if input_names is not None:
         layer['inputs'] = input_names
 
+    layer['data_format'] = keras_layer['config'].get('data_format', 'channels_last')
+
     if 'activation' in keras_layer['config']:
         layer['activation'] = keras_layer['config']['activation']
     if 'epsilon' in keras_layer['config']:
@@ -121,6 +123,72 @@ def parse_default_keras_layer(keras_layer, input_names):
 
     return layer
 
+def parse_data_format(input_shape, data_format='channels_last'):
+    # Ignore batch size
+    input_shape = input_shape[1:]
+    
+    if data_format.lower() == 'channels_last':
+        if len(input_shape) == 2: # 1D, (n_in, n_filt)
+            return (input_shape[0], input_shape[1])
+        elif len(input_shape) == 3: # 2D, (in_height, in_width, n_filt)
+            return (input_shape[0], input_shape[1], input_shape[2])
+        
+    elif data_format.lower() == 'channels_first':
+        if len(input_shape) == 2: # 1D, (n_filt, n_in)
+            return (input_shape[1], input_shape[0])
+        elif len(input_shape) == 3: # 2D, (n_filt, in_height, in_width)
+            return (input_shape[1], input_shape[2], input_shape[0])
+    else:
+        raise Exception('Unknown data format: {}'.format(data_format))
+
+def compute_padding_1d(pad_type, in_size, stride, filt_size):
+    if pad_type.lower() == 'same':
+        n_out = int(math.ceil(float(in_size) / float(stride)))
+        if (in_size % stride == 0):
+            pad_along_size = max(filt_size - stride, 0)
+        else:
+            pad_along_size = max(filt_size - (in_size % stride), 0)
+        pad_left  = pad_along_size // 2
+        pad_right  = pad_along_size - pad_left
+    elif pad_type.lower() == 'valid':
+        n_out = int(math.ceil(float(in_size - filt_size + 1) / float(stride)))
+        pad_left = 0
+        pad_right = 0
+    else:
+        raise Exception('Unknown padding type: {}'.format(pad_type))
+
+    return (n_out, pad_left, pad_right)
+
+def compute_padding_2d(pad_type, in_height, in_width, stride_height, stride_width, filt_height, filt_width):
+    if pad_type.lower() == 'same':
+        #Height
+        out_height = int(math.ceil(float(in_height) / float(stride_height)))
+        if (in_height % stride_height == 0):
+            pad_along_height = max(filt_height - stride_height, 0)
+        else:
+            pad_along_height = max(filt_height - (in_height % stride_height), 0)
+        pad_top = pad_along_height // 2
+        pad_bottom = pad_along_height - pad_top
+        #Width
+        out_width = int(math.ceil(float(in_width) / float(stride_width)))
+        if (in_width % stride_width == 0):
+            pad_along_width = max(filt_width - stride_width, 0)
+        else:
+            pad_along_width = max(filt_width - (in_width % stride_width), 0)
+        pad_left = pad_along_width // 2
+        pad_right = pad_along_width - pad_left
+    elif pad_type.lower() == 'valid':
+        out_height = int(math.ceil(float(in_height - filt_height + 1) / float(stride_height)))
+        out_width = int(math.ceil(float(in_width - filt_width + 1) / float(stride_width)))
+        
+        pad_top = 0
+        pad_bottom = 0
+        pad_left = 0
+        pad_right = 0
+    else:
+        raise Exception('Unknown padding type: {}'.format(pad_type))
+
+    return (out_height, out_width, pad_top, pad_bottom, pad_left, pad_right)
 
 def keras_to_hls(config):
 
@@ -133,8 +201,12 @@ def keras_to_hls(config):
 
     if 'KerasModel' in config:
         # Model instance passed in config from API
-        model_arch = json.loads(config['KerasModel'].to_json())
-        reader = KerasModelReader(config['KerasModel'])
+        keras_model = config['KerasModel']
+        if isinstance(keras_model, str):
+            from tensorflow.keras.models import load_model
+            keras_model = load_model(keras_model)
+        model_arch = json.loads(keras_model.to_json())
+        reader = KerasModelReader(keras_model)
     elif 'KerasJson' in config:
         # Extract model architecture from json
         with open(config['KerasJson']) as json_file:
@@ -156,7 +228,7 @@ def keras_to_hls(config):
     #print(model_arch)
 
     #Define layers to skip for conversion to HLS
-    skip_layers = ['Dropout', 'Flatten']
+    skip_layers = ['Dropout']
     #All supported layers
     supported_layers = get_supported_keras_layers() + skip_layers
 
@@ -200,10 +272,13 @@ def keras_to_hls(config):
     print('Topology:')
     for keras_layer in layer_config:
         if 'batch_input_shape' in keras_layer['config']:
-            input_shapes = [keras_layer['config']['batch_input_shape']]
+            if 'inbound_nodes' in keras_layer and len(keras_layer['inbound_nodes']) > 0:
+                input_shapes = [output_shapes[inbound_node[0]] for inbound_node in keras_layer['inbound_nodes'][0]]
+            else:
+                input_shapes = [keras_layer['config']['batch_input_shape']]
         else:
             if 'inbound_nodes' in keras_layer:
-                input_shapes = [output_shapes[inbound_node[0][0]] for inbound_node in keras_layer['inbound_nodes']]
+                input_shapes = [output_shapes[inbound_node[0]] for inbound_node in keras_layer['inbound_nodes'][0]]
             else:
                 # Sequential model, so output_shape from the previous layer is still valid 
                 input_shapes = [output_shape]
@@ -218,10 +293,7 @@ def keras_to_hls(config):
                 #Skipped layers can follow each other (e.g., Dropout -> Flatten)
                 inputs_map[name] = inputs_map.get(parent_input, parent_input)
 
-            if keras_class == 'Flatten':
-                output_shapes[keras_layer['config']['name']] = [input_shapes[0][0], np.prod(input_shapes[0][1:])]
-            else:
-                output_shapes[keras_layer['config']['name']] = input_shapes[0]
+            output_shapes[keras_layer['config']['name']] = input_shapes[0]
 
             continue
 
@@ -236,9 +308,9 @@ def keras_to_hls(config):
 
         layer, output_shape = layer_handlers[keras_class](keras_layer, input_names, input_shapes, reader, config)
 
-        print('Layer name: {}, layer type: {}, current shape: {}'.format(layer['name'], layer['class_name'], input_shapes))
+        print('Layer name: {}, layer type: {}, input shapes: {}, output shape: {}'.format(layer['name'], layer['class_name'], input_shapes, output_shape))
         layer_list.append( layer )
-        if 'activation' in layer and layer['class_name'] not in ['Activation', 'LeakyReLU', 'ThresholdedReLU', 'ELU', 'PReLU', 'Softmax']:# + qkeras_layers:
+        if 'activation' in layer and layer['class_name'] not in ['Activation', 'LeakyReLU', 'ThresholdedReLU', 'ELU', 'PReLU', 'Softmax', 'TernaryTanh']:# + qkeras_layers:
             act_layer = {}
             act_layer['name'] = layer['name'] + '_' + layer['activation']
             act_layer['activation'] = layer['activation']
@@ -247,6 +319,7 @@ def keras_to_hls(config):
                 act_layer['class_name'] = layer['activation']
             elif layer['activation'] == 'softmax':
                 act_layer['class_name'] = 'Softmax'
+                act_layer['axis'] = -1
             else:
                 act_layer['class_name'] = 'Activation'
             inputs_map[layer['name']] = act_layer['name']
