@@ -13,11 +13,11 @@ from hls4ml.model.hls_layers import *
 from hls4ml.templates import get_backend
 from hls4ml.writer import get_writer
 from hls4ml.model.optimizer import optimize_model, get_available_passes
+from hls4ml.report.vivado_report import parse_vivado_report
 
 class HLSConfig(object):
     def __init__(self, config):
         self.config = config
-
         self.backend = get_backend(self.config.get('Backend', 'Vivado'))
         self.writer = get_writer(self.config.get('Backend', 'Vivado'))
 
@@ -29,10 +29,18 @@ class HLSConfig(object):
         self.layer_type_rf = {}
         self.layer_name_rf = {}
 
+        self.model_targ_cycles = None
+        self.layer_type_targ_cycles = {}
+        self.layer_name_targ_cycles = {}
+
         self.model_strategy = 'Latency'
         self.layer_type_strategy = {}
         self.layer_name_strategy = {}
 
+        self.model_conv_implementation = 'LineBuffer'
+        self.layer_type_conv_implementation = {}
+        self.layer_name_conv_implementation = {}
+
         self.model_compression = False
         self.layer_type_compression = {}
         self.layer_name_compression = {}
@@ -106,8 +114,14 @@ class HLSConfig(object):
         if precision is None:
             raise Exception('No precision for {}->{} found and no default specified.'.format(layer.name, var))
 
+        precision = self.backend.convert_precision_string(precision)
+
         return (precision, type_name)
 
+    def get_bram_size(self, layer):
+        bf = self.model_bf
+        return bf
+
     def get_reuse_factor(self, layer):
         rf = self.layer_name_rf.get(layer.name.lower())
         if rf is None:
@@ -120,6 +134,15 @@ class HLSConfig(object):
 
         return rf
 
+    def get_target_cycles(self, layer):
+        targ_cycles = self.layer_name_targ_cycles.get(layer.name.lower())
+        if targ_cycles is None:
+            targ_cycles = self.layer_name_targ_cycles.get(layer.__class__.__name__.lower())
+        if targ_cycles is None:
+            targ_cycles = self.model_targ_cycles
+ 
+        return targ_cycles
+
     def get_strategy(self, layer):
         strategy = self.layer_name_strategy.get(layer.name.lower())
         if strategy is None:
@@ -128,6 +151,15 @@ class HLSConfig(object):
             strategy = self.model_strategy
 
         return strategy
+    
+    def get_conv_implementation(self, layer):
+        conv_implementation = self.layer_name_conv_implementation.get(layer.name.lower())
+        if conv_implementation is None:
+            conv_implementation = self.layer_type_conv_implementation.get(layer.__class__.__name__.lower())
+        if conv_implementation is None:
+            conv_implementation = self.model_conv_implementation
+
+        return conv_implementation
 
     def is_resource_strategy(self, layer):
         return self.get_strategy(layer).lower() == 'resource'
@@ -167,7 +199,10 @@ class HLSConfig(object):
                 else:
                     self.model_precision['default'] = precision_cfg # Default precision for everything
 
+            self.model_bf = model_cfg.get('BramFactor', np.inf) # Weight threshold to be external BRAM
             self.model_rf = model_cfg.get('ReuseFactor')
+            self.model_targ_cycles = model_cfg.get('TargetCycles')
+            self.model_conv_implementation = model_cfg.get('ConvImplementation', 'LineBuffer')
             self.model_strategy = model_cfg.get('Strategy', 'Latency')
             self.model_compression = bool(model_cfg.get('Compression', 0))
 
@@ -184,11 +219,19 @@ class HLSConfig(object):
                 rf = layer_cfg.get('ReuseFactor')
                 if rf is not None:
                     self.layer_type_rf[layer_type.lower()] = rf
+                
+                targ_cycles = layer_cfg.get('TargetCycles')
+                if targ_cycles is not None:
+                    self.layer_type_targ_cycles[layer_type.lower()] = targ_cycles
 
                 strategy = layer_cfg.get('Strategy')
                 if strategy is not None:
                     self.layer_type_strategy[layer_type.lower()] = strategy
 
+                conv_implementation = layer_cfg.get('ConvImplementation')
+                if conv_implementation is not None:
+                    self.layer_type_conv_implementation[layer_type.lower()] = conv_implementation
+
                 compression = layer_cfg.get('Compression')
                 if compression is not None:
                     self.layer_type_compression[layer_type.lower()] = bool(compression)
@@ -207,10 +250,18 @@ class HLSConfig(object):
                 if rf is not None:
                     self.layer_name_rf[layer_name.lower()] = rf
 
+                targ_cycles = layer_cfg.get('TargetCycles')
+                if targ_cycles is not None:
+                    self.layer_name_targ_cycles[layer_name.lower()] = targ_cycles
+
                 strategy = layer_cfg.get('Strategy')
                 if strategy is not None:
                     self.layer_name_strategy[layer_name.lower()] = strategy
 
+                conv_implementation = layer_cfg.get('ConvImplementation')
+                if conv_implementation is not None:
+                    self.layer_name_conv_implementation[layer_name.lower()] = conv_implementation
+
                 compression = layer_cfg.get('Compression')
                 if compression is not None:
                     self.layer_name_compression[layer_name.lower()] = bool(compression)
@@ -257,6 +308,9 @@ class HLSModel(object):
         self.graph = OrderedDict()
         self.output_vars = {}
 
+        # External BRAM 
+        self.bram_vars = {}
+
         self._top_function_lib = None
 
         self._make_graph(layer_list)
@@ -269,7 +323,9 @@ class HLSModel(object):
             name = layer['name']
             inputs = layer.get('inputs', [])
             outputs = layer.get('outputs', [])
-            if len(inputs) == 0:
+            if kind in ['InputLayer', 'Input']:
+                inputs = ['input']
+            elif len(inputs) == 0:
                 inputs = [next(reversed(self.graph), 'input')]
             if len(outputs) == 0:
                 outputs = [name]
@@ -280,21 +336,66 @@ class HLSModel(object):
         optimize_model(self, optimizers)
 
     def make_node(self, kind, name, attributes, inputs, outputs=None):
+        """ Make a new node not connected to the model graph.
+
+        The 'kind' should be a valid layer registered with `register_layer`. If no outputs
+        are specified, a default output named the same as the node will be created. The 
+        returned node should be added to the graph with `insert_node` or `replace_node`
+        functions.
+
+        Args:
+            kind (str): Type of node to add
+            name (str): Name of the node
+            attributes (dict): Initial set of attributes required to construct the node (Layer)
+            inputs (list): List of inputs to the layer
+            outputs (list, optional): The optional list of named outputs of the node
+
+        Raises:
+            Exception: If an attempt to insert a node with multiple inputs is made or if
+                `before` does not specify a correct node in sequence.
+
+        Returns:
+            Layer: The node created.
+        """
+
+        if kind not in layer_map:
+            raise Exception('Layer {} not found in registry.'.format(kind))
+
         node = layer_map[kind](self, name, attributes, inputs, outputs)
         for o in node.outputs:
             out_var = node.get_output_variable(output_name=o)
             if o in self.outputs:
                 out_var.type.name = 'result_t'
             self.output_vars[o] = out_var
-
         return node
 
-    def insert_node(self, node):
+    def insert_node(self, node, before=None):
+        """ Insert a new node into the model graph.
+
+        The node to be inserted should be created with `make_node()` function. The optional 
+        parameter `before` can be used to specify the node that follows in case of ambiguities.
+
+        Args:
+            node (Layer): Node to insert
+            before (Layer, optional): The next node in sequence before which a
+                new node should be inserted. 
+        Raises:
+            Exception: If an attempt to insert a node with multiple inputs is made or if
+                `before` does not specify a correct node in sequence.
+
+        """
         if len(node.inputs) > 1:
             raise Exception('Cannot insert a node with more than one input (for now).')
 
-        prev_node = self.graph.get(node.inputs[0])
-        next_node = next((x for x in self.graph.values() if x.inputs[0] == prev_node.outputs[0]), None)
+        prev_node = node.get_input_node(node.inputs[0])
+        next_nodes = [x for x in self.graph.values() if x.inputs[0] in prev_node.outputs]
+        if before is None:
+            next_node = next((x for x in self.graph.values() if x.inputs[0] in prev_node.outputs), None)
+        else:
+            if before not in next_nodes:
+                raise Exception('Cannot insert a node {} before {} (candidates: {}).'.format(node.name, before.name, ','.join([n.name for n in next_nodes])))
+            next_node = before
+
         if next_node is not None:
             next_node.inputs[0] = node.outputs[0]
 
@@ -305,28 +406,53 @@ class HLSModel(object):
                 new_graph[node.name] = node
 
         self.graph = new_graph
+        self._update_model_outputs()
 
     def remove_node(self, node, rewire=True):
+        """ Remove a node from a graph.
+
+        By default, this function can connect the outputs of previous node to the input of next one.
+        Note that when removing a leaf node `rewire` should be set to `False`.
+
+        Args:
+            node (Layer): The node to remove
+            rewire (bool, optional): If `True`, connects the outputs of the previous node
+                to the inputs of the next node
+
+        Raises:
+            Exception: If an attempt is made to rewire a leaf node or a node with multiple
+                inputs/outpus.
+
+        """
         if rewire:
             if len(node.inputs) > 1 or len(node.outputs) > 1:
                 raise Exception('Cannot rewire a node with multiple inputs/outputs')
             prev_node = self.graph.get(node.inputs[0])
-            next_node = next((x for x in self.graph.values() if x.inputs[0] == node.outputs[0]), None)
+            next_node = next((x for x in self.graph.values() if node.outputs[0] in x.inputs), None)
             if prev_node is not None:
                 if next_node is not None:
-                    next_node.inputs[0] = prev_node.outputs[0]
+                    for i,_ in enumerate(next_node.inputs):
+                        if node.outputs[0] == next_node.inputs[i]:
+                            next_node.inputs[i] = prev_node.outputs[0]
+                            break
                 else:
-                    if node.outputs[0] in self.outputs:
-                        self.outputs = [prev_node.outputs[0] if x == node.outputs[0] else x for x in self.outputs]
-                    else:
+                    if not node.outputs[0] in self.outputs:
                         raise Exception('Cannot rewire a node without child')
             else:
                 raise Exception('Cannot rewire a node without a parent')
-
+        
         del self.output_vars[node.outputs[0]]
         del self.graph[node.name]
+        self._update_model_outputs()
 
     def replace_node(self, old_node, new_node):
+        """ Replace an existing node in the graph with a new one.
+
+        Args:
+            old_node (Layer): The node to replace
+            new_node (Layer): The new node
+
+        """
         prev_node = self.graph.get(old_node.inputs[0])
         next_node = next((x for x in self.graph.values() if x.inputs[0] == old_node.outputs[0]), None)
         if next_node is not None:
@@ -336,6 +462,18 @@ class HLSModel(object):
                 new_node.inputs = [prev_node.outputs[0]]
 
         self.graph = OrderedDict((new_node.name, new_node) if k == old_node.name else (k, v) for k, v in self.graph.items())
+        self._update_model_outputs()
+
+    def _update_model_outputs(self):
+        ''' Update the model outputs
+
+        All node outputs and inputs are found. The model outputs are set to all node outputs
+        that are not also node inputs.
+        '''
+        node_outputs = np.array([out for node in self.graph.values() for out in node.outputs])
+        node_inputs = np.array([inp for node in self.graph.values() for inp in node.inputs])
+        model_outputs = node_outputs[np.isin(node_outputs, node_inputs, invert=True)]
+        self.outputs = model_outputs.tolist()
 
     def get_weights_data(self, layer_name, var_name):
         return self.reader.get_weights_data(layer_name, var_name)
@@ -353,6 +491,12 @@ class HLSModel(object):
             variables.append(self.graph[inp].get_output_variable())
         return variables
 
+    def register_bram_variable(self, out_name, variable):
+        self.bram_vars[out_name] = variable
+
+    def get_bram_variables(self):
+        return self.bram_vars.values()
+
     def register_output_variable(self, out_name, variable):
         if out_name in self.outputs:
             variable.type.name = 'result_t'
@@ -365,7 +509,7 @@ class HLSModel(object):
         return variables
 
     def get_layer_output_variable(self, output_name):
-        return self.output_vars[output_name]
+        return self.output_vars.get(output_name, None)
 
     def write(self):
         def make_stamp():
@@ -406,55 +550,80 @@ class HLSModel(object):
     def _get_top_function(self, x):
         if self._top_function_lib is None:
             raise Exception('Model not compiled')
-        if len(self.get_input_variables()) > 1 or len(self.get_input_variables()) > 1:
-            raise Exception('Calling "predict" on models with multiple inputs or outputs is not supported (yet)')
+        if len(self.get_input_variables()) == 1:
+            xlist = [x]
+        else: 
+            xlist = x
+        
+        for xi in xlist:
+            if not isinstance(xi, np.ndarray):
+                raise Exception('Expected numpy.ndarray, but got {}'.format(type(x)))
+            if not xi.flags['C_CONTIGUOUS']:
+                raise Exception('Array must be c_contiguous, try using numpy.ascontiguousarray(x)')
 
-        if not isinstance(x, np.ndarray):
-            raise Exception('Expected numpy.ndarray, but got {}'.format(type(x)))
-        if not x.flags['C_CONTIGUOUS']:
-            raise Exception('Array must be c_contiguous, try using numpy.ascontiguousarray(x)')
-
-        if x.dtype in [np.single, np.float32]:
+        x0 = xlist[0]
+        if x0.dtype in [np.single, np.float32]:
             top_function = getattr(self._top_function_lib, self.config.get_project_name() + '_float')
             ctype = ctypes.c_float
-        elif x.dtype in [np.double, np.float64, np.float_]:
+        elif x0.dtype in [np.double, np.float64, np.float_]:
             top_function = getattr(self._top_function_lib, self.config.get_project_name() + '_double')
             ctype = ctypes.c_double
         else:
-            raise Exception('Invalid type ({}) of numpy array. Supported types are: single, float32, double, float64, float_.'.format(x.dtype))
+            raise Exception('Invalid type ({}) of numpy array. Supported types are: single, float32, double, float64, float_.'.format(x0.dtype))
+
 
         top_function.restype = None
-        top_function.argtypes = [npc.ndpointer(ctype, flags="C_CONTIGUOUS"), npc.ndpointer(ctype, flags="C_CONTIGUOUS"),
-            ctypes.POINTER(ctypes.c_ushort), ctypes.POINTER(ctypes.c_ushort)]
+        top_function.argtypes = [npc.ndpointer(ctype, flags="C_CONTIGUOUS") for i in range(len(xlist)+1)]
+        top_function.argtypes += [ctypes.POINTER(ctypes.c_ushort) for i in range(len(xlist)+1)]
 
         return top_function, ctype
 
     def _compute_n_samples(self, x):
-        expected_size = self.get_input_variables()[0].size()
-        x_size = np.prod(x.shape)
-        n_samples, rem = divmod(x_size, expected_size)
-        if rem != 0:
-            raise Exception('Input size mismatch, got {}, expected {}'.format(x_size.shape, self.get_input_variables()[0].shape))
+        if len(self.get_input_variables()) == 1:
+            xlist = [x]
+        else:
+            xlist = x
+        n_samples = []
+        for i, xi in enumerate(xlist):
+            expected_size = self.get_input_variables()[i].size()
+            x_size = np.prod(xi.shape)
+            n_sample, rem = divmod(x_size, expected_size)
+            if rem != 0:
+                raise Exception('Input size mismatch, got {}, expected {}'.format(x_size.shape, self.get_input_variables()[i].shape))
+            n_samples.append(n_sample)
 
-        return n_samples
+        if not all([n_samples[i] == n_samples[i+1] for i in range(len(xlist)-1)]):
+            raise Exception('Input size mismatch, not all inputs match')
+
+        return n_sample
 
     def predict(self, x):
         top_function, ctype = self._get_top_function(x)
         n_samples = self._compute_n_samples(x)
+        n_inputs = len(self.get_input_variables())
 
         curr_dir = os.getcwd()
         os.chdir(self.config.get_output_dir() + '/firmware')
 
         output = []
-        if n_samples == 1:
+        if n_samples == 1 and n_inputs == 1:
             x = [x]
 
         try:
             for i in range(n_samples):
                 predictions = np.zeros(self.get_output_variables()[0].size(), dtype=ctype)
-                top_function(x[i], predictions, ctypes.byref(ctypes.c_ushort()), ctypes.byref(ctypes.c_ushort()))
+                if n_inputs == 1:
+                    top_function(x[i], predictions, ctypes.byref(ctypes.c_ushort()), ctypes.byref(ctypes.c_ushort()))
+                else:
+                    inp = [xj[i] for xj in x]
+                    argtuple = inp
+                    argtuple += [predictions]
+                    argtuple += [ctypes.byref(ctypes.c_ushort()) for k in range(len(inp)+1)]
+                    argtuple = tuple(argtuple)
+                    top_function(*argtuple)
                 output.append(predictions)
 
+
             #Convert to numpy array
             output = np.asarray(output)
         finally:
@@ -472,6 +641,7 @@ class HLSModel(object):
 
         top_function, ctype = self._get_top_function(x)
         n_samples = self._compute_n_samples(x)
+        n_inputs = len(self.get_input_variables())
 
         class TraceData(ctypes.Structure):
             _fields_ = [('name', ctypes.c_char_p),
@@ -481,7 +651,7 @@ class HLSModel(object):
         layer_sizes = {}
         n_traced = 0
         for layer in self.get_layers():
-            if layer.function_cpp() and self.config.get_layer_config_value(layer, 'Trace', False):
+            if layer.function_cpp() and layer.get_attr('Trace', False):
                 n_traced += len(layer.get_variables())
                 trace_output[layer.name] = []
                 layer_sizes[layer.name] = layer.get_output_variable().shape
@@ -503,7 +673,7 @@ class HLSModel(object):
         os.chdir(self.config.get_output_dir() + '/firmware')
 
         output = []
-        if n_samples == 1:
+        if n_samples == 1 and n_inputs == 1:
             x = [x]
 
         try:
@@ -511,7 +681,15 @@ class HLSModel(object):
 
             for i in range(n_samples):
                 predictions = np.zeros(self.get_output_variables()[0].size(), dtype=ctype)
-                top_function(x[i], predictions, ctypes.byref(ctypes.c_ushort()), ctypes.byref(ctypes.c_ushort()))
+                if n_inputs == 1:
+                    top_function(x[i], predictions, ctypes.byref(ctypes.c_ushort()), ctypes.byref(ctypes.c_ushort()))
+                else:
+                    inp = [xj[i] for xj in x]
+                    argtuple = inp
+                    argtuple += [predictions]
+                    argtuple += [ctypes.byref(ctypes.c_ushort()) for k in range(len(inp)+1)]
+                    argtuple = tuple(argtuple)
+                    top_function(*argtuple)
                 output.append(predictions)
                 collect_func(trace_data)
                 for trace in trace_data:
@@ -538,9 +716,9 @@ class HLSModel(object):
     def build(self, reset=False, csim=True, synth=True, cosim=False, validation=False, export=False, vsynth=False):
         if 'linux' in sys.platform:
             backend = self.config.get_config_value('Backend', 'Vivado')
-            if backend == 'Vivado':
+            if backend in ['Vivado', 'VivadoAccelerator']:
                 found = os.system('command -v vivado_hls > /dev/null')
-                if found is not 0:
+                if found != 0:
                     raise Exception('Vivado HLS installation not found. Make sure "vivado_hls" is on PATH.')
 
             elif backend == 'Intel':
@@ -560,3 +738,5 @@ class HLSModel(object):
             .format(reset=reset, csim=csim, synth=synth, cosim=cosim, validation=validation, export=export, vsynth=vsynth))
         os.chdir(curr_dir)
 
+        return parse_vivado_report(self.config.get_output_dir())
+
