@@ -2,7 +2,8 @@ import numpy as np
 import re
 
 from hls4ml.model.optimizer import OptimizerPass
-from hls4ml.model.hls_model import Layer, IntegerPrecisionType, register_layer
+from hls4ml.model.hls_model import Layer, IntegerPrecisionType, XnorPrecisionType, register_layer
+from hls4ml.model.hls_layers import BatchNormalization
 from hls4ml.templates import templates
 
 class BatchNormalizationQuantizedTanh(Layer):
@@ -15,34 +16,12 @@ class BatchNormalizationQuantizedTanh(Layer):
         inp = self.get_input_variable()
         shape = inp.shape
         dims = inp.dim_names
-        precision_bits = re.search('.+<(.+?)>', inp.type.precision).group(1).split(',')
-        if 'int' in str(inp.type.precision):
-            W = int(precision_bits[0])
-            I = W
-            F = 0
-        elif 'fixed' in str(inp.type.precision):
-            W = int(precision_bits[0])
-            I = int(precision_bits[1])
-            F = W - I
-        original_name = self.attributes.get('original_name')
-        variance = self.model.get_weights_data(original_name, 'moving_variance')
-        mean = self.model.get_weights_data(original_name, 'moving_mean')
-        gamma = self.model.get_weights_data(original_name, 'gamma')
-        beta = self.model.get_weights_data(original_name, 'beta')
-        epsilon = self.attributes.get('epsilon')
-        threshold = mean - beta * np.sqrt(variance + epsilon) / gamma
         if self.get_attr('quantize') == 2:
-            self.add_output_variable(shape, dims, precision=IntegerPrecisionType(width=1, signed=False))
-            threshold = np.floor(threshold * 2**F) / 2**F
-            self.add_weights_variable(name='threshold', var_name='t{index}', data=threshold, type_name='threshold{index}_t', precision=inp.type.precision)
+            self.add_output_variable(shape, dims, precision=XnorPrecisionType())
         elif self.get_attr('quantize') == 3:
             self.add_output_variable(shape, dims, precision=IntegerPrecisionType(width=2))
-            threshold_hi = 0.5/(gamma/np.sqrt(variance + epsilon)) + threshold
-            threshold_lo = -0.5/(gamma/np.sqrt(variance + epsilon)) + threshold
-            threshold_hi = np.floor(threshold_hi * 2**F) / 2**F
-            threshold_lo = np.floor(threshold_lo * 2**F) / 2**F
-            self.add_weights_variable(name='threshold_hi', var_name='th{index}', data=threshold_hi, type_name='threshold_hi_{index}_t', precision=inp.type.precision)
-            self.add_weights_variable(name='threshold_lo', var_name='tl{index}', data=threshold_lo, type_name='threshold_lo_{index}_t', precision=inp.type.precision)
+        else:
+            raise Exception('Unsupported quantize attribute for BatchNormalizationQuantizedTanh: {}'.format(self.get_attr('quantize')))
 
     def function_cpp(self):
         params = self._default_function_params()
@@ -61,6 +40,26 @@ class BatchNormalizationQuantizedTanh(Layer):
 
         return self._config_template.format(**params)
 
+    def set_thresholds(self, scale, bias, ternary_threshold=0.5):
+        inp = self.get_input_variable()
+        shape = inp.shape
+        dims = inp.dim_names
+        precision = self.model.config.backend.convert_precision_string(inp.type.precision)
+        W, I, F = precision.width, precision.integer, precision.fractional
+        threshold = - bias / scale
+        if self.get_attr('quantize') == 2:
+            self.add_output_variable(shape, dims, precision=XnorPrecisionType())
+            threshold = np.floor(threshold * 2**F) / 2**F
+            self.add_weights_variable(name='threshold', var_name='t{index}', data=threshold, type_name='threshold{index}_t', precision=inp.type.precision)
+        elif self.get_attr('quantize') == 3:
+            self.add_output_variable(shape, dims, precision=IntegerPrecisionType(width=2))
+            threshold_hi = ternary_threshold / scale + threshold
+            threshold_lo = -ternary_threshold / scale + threshold
+            threshold_hi = np.floor(threshold_hi * 2**F) / 2**F
+            threshold_lo = np.floor(threshold_lo * 2**F) / 2**F
+            self.add_weights_variable(name='threshold_hi', var_name='th{index}', data=threshold_hi, type_name='threshold_hi_{index}_t', precision=inp.type.precision)
+            self.add_weights_variable(name='threshold_lo', var_name='tl{index}', data=threshold_lo, type_name='threshold_lo_{index}_t', precision=inp.type.precision)
+
 batchnorm_quantized_tanh_config_template = """struct config{index} : nnet::batchnorm_quantized_tanh_config {{
     static const unsigned n_in = {n_in};
     static const unsigned n_filt = {n_filt};
@@ -73,25 +72,27 @@ batchnorm_quantized_tanh_function_template = 'nnet::normalize_{quantize}_tanh<{i
 # Register the layer types to the layer map
 register_layer('BatchNormalizationQuantizedTanh', BatchNormalizationQuantizedTanh)
 
+from hls4ml.templates.vivado_template import batchnorm_include_list
+
 # Register the templates for config and function
-templates.get_backend('Vivado').register_templates('BatchNormalizationQuantizedTanh', batchnorm_quantized_tanh_function_template, batchnorm_quantized_tanh_config_template)
+for backend in ['Vivado', 'VivadoAccelerator']:
+    templates.get_backend(backend).register_templates('BatchNormalizationQuantizedTanh', batchnorm_quantized_tanh_function_template, batchnorm_quantized_tanh_config_template, batchnorm_include_list)
 
 class MergeBatchNormAndQuantizedTanh(OptimizerPass):
     def match(self, node):
         is_match = (node.__class__.__name__ == 'Activation'
-            and node.get_attr('activation') in ['binary_tanh', 'ternary_tanh']
-            and node.get_input_node().__class__.__name__ == 'BatchNormalization')
+            and node.get_attr('activation') in ['binary', 'binary_tanh', 'ternary', 'ternary_tanh']
+            or node.__class__.__name__ == 'TernaryTanh')
+        is_match = is_match and isinstance(node.get_input_node(), BatchNormalization)
         return is_match
 
     def transform(self, model, node):
         bn_layer = node.get_input_node()
-        # Remove the Activation layer
-        model.remove_node(node, rewire=True)
         # Make a new layer with the new attributes
         quantize = 0
-        if node.get_attr('activation') == 'binary_tanh':
+        if 'binary' in node.get_attr('activation'):
             quantize = 2
-        if node.get_attr('activation') == 'ternary_tanh':
+        if 'ternary' in node.get_attr('activation'):
             quantize = 3
         attrs = {
             'name' : bn_layer.get_attr('name'),
@@ -100,19 +101,24 @@ class MergeBatchNormAndQuantizedTanh(OptimizerPass):
             'n_in' : bn_layer.get_attr('n_in'),
             'n_out' : bn_layer.get_attr('n_in'),
             'n_filt' : bn_layer.get_attr('n_filt'),
-            'epsilon' : bn_layer.get_attr('epsilon'),
-            'quantize' : quantize
+            'quantize' : quantize,
+            'Trace' : bn_layer.get_attr('Trace')
         }
         bnbt_layer = model.make_node('BatchNormalizationQuantizedTanh', 'bnbt_' + bn_layer.name, attrs, bn_layer.inputs)
-        # Replace the old BatchNormalization layer with this one
-        model.replace_node(bn_layer, bnbt_layer)
+        bnbt_layer.set_thresholds(bn_layer.get_weights('scale').data, bn_layer.get_weights('bias').data, node.get_attr('threshold',0.5))
+        # Remove the BatchNormalization layer
+        model.remove_node(bn_layer, rewire=True)
+        # Replace the old Activation layer with this one
+        model.replace_node(node, bnbt_layer)
 
         return True
 
 class QuantizeDenseOutput(OptimizerPass):
     def match(self, node):
-        is_match = (node.__class__.__name__ == 'Dense' and node.get_attr('weight_quantizer') is not None
-            and node.get_input_node().__class__.__name__ == 'BatchNormalizationQuantizedTanh')
+        is_match = node.__class__.__name__ == 'Dense'
+        is_match = is_match and node.get_input_node().__class__.__name__ == 'BatchNormalizationQuantizedTanh'
+        quantizer = node.get_attr('weight_quantizer')
+        is_match = is_match and (quantizer.__class__.__name__ == 'BinaryQuantizer' or quantizer.__class__.__name__ == 'TernaryQuantizer')
         return is_match
 
     def transform(self, model, node):
@@ -129,7 +135,7 @@ class QuantizeDenseOutput(OptimizerPass):
         quantized_precision = None
         quantizer = node.get_attr('weight_quantizer')
         if quantizer.__class__.__name__ == 'BinaryQuantizer':
-            quantized_precision = IntegerPrecisionType(width=1, signed=False)
+            quantized_precision = XnorPrecisionType()
         elif quantizer.__class__.__name__ == 'TernaryQuantizer':
             quantized_precision = IntegerPrecisionType(width=2)
         else:
@@ -167,3 +173,4 @@ class QuantizeDenseOutput(OptimizerPass):
                     threshold_var.data = np.floor(threshold_var.data)
 
         return False
+
