@@ -3,6 +3,9 @@
 
 #include "ap_int.h"
 #include "ap_fixed.h"
+#include "nnet_utils/nnet_types.h"
+#include <cstddef>
+#include <cstdio>
 
 //hls-fpga-machine-learning insert numbers
 
