@@ -26,6 +26,9 @@
 #include <math.h>
 
 #include "firmware/myproject.h"
+#include "firmware/nnet_utils/nnet_helpers.h"
+
+//hls-fpga-machine-learning insert bram
 
 #define CHECKPOINT 5000
 
@@ -76,8 +79,6 @@ int main(int argc, char **argv)
 
       //hls-fpga-machine-learning insert top-level-function
 
-      //hls-fpga-machine-learning insert tb-output
-
       if (e % CHECKPOINT == 0) {
         std::cout << "Predictions" << std::endl;
         //hls-fpga-machine-learning insert predictions
@@ -85,11 +86,15 @@ int main(int argc, char **argv)
         //hls-fpga-machine-learning insert quantized
       }
       e++;
+
+      //hls-fpga-machine-learning insert tb-output
+
     }
     fin.close();
     fpr.close();
   } else {
     std::cout << "INFO: Unable to open input/predictions file, using default input." << std::endl;
+
     //hls-fpga-machine-learning insert zero
 
     //hls-fpga-machine-learning insert top-level-function
@@ -97,6 +102,7 @@ int main(int argc, char **argv)
     //hls-fpga-machine-learning insert output
 
     //hls-fpga-machine-learning insert tb-output
+
   }
 
   fout.close();
