@@ -25,30 +25,13 @@
 // This is a substitute for "ceil(n/(float)d)".
 #define DIV_ROUNDUP(n,d) ((n + d - 1) / d)
 #define MIN(n,d) (n > d ? d : n)
+#define MAX(n,d) (n > d ? n : d)
 
 namespace nnet {
 
 // Common type definitions
-enum io_type {io_parallel = 0, io_serial};
-
-// Default data types (??) TODO: Deprecate
-typedef ap_fixed<16,4>  weight_t_def;
-typedef ap_fixed<16,4>  bias_t_def;
-typedef ap_fixed<32,10> accum_t_def;
-
- template<class data_T, int NIN1, int NIN2>
-   void merge(
-	      data_T data1[NIN1], 
-	      data_T data2[NIN2],
-	      data_T res[NIN1+NIN2])
- {
-   for(int ii=0; ii<NIN1; ii++){
-     res[ii] = data1[ii];
-   }
-   for(int ii=0; ii<NIN2; ii++){
-     res[NIN1+ii] = data2[ii];
-   }
- }
+enum io_type {io_parallel = 0, io_serial, io_stream};
+enum strategy { latency, resource };
 
  /* ---
   * Balanced tree reduce implementation.
@@ -59,27 +42,18 @@ typedef ap_fixed<32,10> accum_t_def;
   * before applying and accumulate the result over the rolled dimension.
   * --- */
  template<class T, int N, class Op>
- T reduce(T* x, Op op){
-	static constexpr int leftN = pow2(floorlog2(N - 1)) > 0 ? pow2(floorlog2(N - 1)) : 0;
-	static constexpr int rightN = N - leftN > 0 ? N - leftN : 0;
-	if(N == 1){
-		return x[0];
-	}else if(N == 2){
-		return op(x[0],x[1]);
-	}else{
-		T left[leftN];
-		T right[rightN];
-		#pragma HLS array_partition variable=left complete
-		#pragma HLS array_partition variable=right complete
-		for(int i = 0; i < leftN; i++){
-			left[i] = x[i];
-		}
-		for(int i = 0; i < rightN; i++){
-			right[i] = x[i+leftN];
-		}
-		return op(reduce<T,leftN,Op>(left, op), reduce<T,rightN,Op>(right, op));
-	}
- }
+ T reduce(const T* x, Op op)
+ {
+     static constexpr int leftN = pow2(floorlog2(N - 1)) > 0 ? pow2(floorlog2(N - 1)) : 0;
+     static constexpr int rightN = N - leftN > 0 ? N - leftN : 0;
+     if (N == 1){
+         return x[0];
+     }
+     if (N == 2){
+         return op(x[0],x[1]);
+     }
+     return op(reduce<T,leftN,Op>(x, op), reduce<T,rightN,Op>(x+leftN, op));
+ } 
 
  template<class T>
  class Op_add{
