@@ -1,22 +1,25 @@
 from __future__ import print_function
-import numpy as np
-import h5py
 import json
-import math
-from collections import OrderedDict
 
-def create_vivado_config(output_dir='my-hls-test', project_name='myproject',
-    fpga_part='xcku115-flvb2104-2-i', clock_period=5):
-    
+import hls4ml
+
+
+def create_config(output_dir='my-hls-test', project_name='myproject',
+    backend='Vivado', **kwargs):
+
+    backend_list = hls4ml.templates.get_available_backends()
+    if backend not in backend_list:
+        raise Exception('Unknown backend: {}'.format(backend))
+
+    backend = hls4ml.templates.get_backend(backend)
+
+    backend_config = backend.create_initial_config(**kwargs)
+
     config = {}
-    
     config['OutputDir'] = output_dir
     config['ProjectName'] = project_name
-    config['XilinxPart'] = fpga_part
-    config['ClockPeriod'] = clock_period
-    config['Backend'] = 'Vivado'
-    config['IOType'] = 'io_parallel' # To become obsolete in the future
-    config['HLSConfig'] = {}
+    config['Backend'] = backend.name
+    config.update(backend_config)
 
     return config
 
@@ -33,11 +36,15 @@ def _get_precision_from_quantizer(quantizer):
         else: 
             quantizer['class_name'] = quantizer_obj.__name__
 
-    supported_quantizers = ['quantized_bits', 'quantized_relu', 'quantized_tanh']
+    supported_quantizers = ['quantized_bits', 'quantized_relu', 'quantized_tanh', 'quantized_po2', 'quantized_relu_po2']
+    signed = True
     if quantizer['class_name'] in supported_quantizers:
-        bits = int(quantizer['config']['bits']) + 1
-        integer = int(quantizer['config']['integer']) + 1
-        
+        bits = int(quantizer['config']['bits'])
+        # if integer isn't specified, it should be the same as bits
+        integer = int(quantizer['config'].get('integer', bits-1)) + 1
+        if quantizer['class_name'] == 'quantized_relu':
+            signed = False
+            integer -= 1
     elif quantizer['class_name'] in ['binary', 'stochastic_binary', 'binary_tanh']:
         bits = 2
         integer = 2
@@ -49,12 +56,41 @@ def _get_precision_from_quantizer(quantizer):
         raise Exception('ERROR: Unsupported quantizer: {}'.format(quantizer['class_name']))
 
     decimal = bits - integer
+    signed = '' if signed else 'u'
     if decimal > 0:
-        return 'ap_fixed<{},{}>'.format(bits, integer)
+        return 'ap_{}fixed<{},{}>'.format(signed, bits, integer)
     else:
-        return 'ap_int<{}>'.format(bits)
+        return 'ap_{}int<{}>'.format(signed, bits)
 
 def config_from_keras_model(model, granularity='model', default_precision='ap_fixed<16,6>', default_reuse_factor=1):
+    """Create an HLS conversion config given the Keras model.
+
+    This function serves as the initial step in creating the custom conversion configuration.
+    Users are advised to inspect the returned object to tweak the conversion configuration.
+    The return object can be passed as `hls_config` parameter to `convert_from_keras_model`.
+
+    Args:
+        model: Keras model
+        granularity (str, optional): Granularity of the created config. Defaults to 'model'.
+            Can be set to 'model', 'type' and 'layer'.
+
+            Granularity can be used to generate a more verbose config that can be fine-tuned.
+            The default granulrity ('model') will generate config keys that apply to the whole
+            model, so changes to the keys will affect the entire model. 'type' granularity will
+            generate config keys that affect all layers of a given type, while the 'name' granularity
+            will generate config keys for every layer separately, allowing for highly specific
+            configuration tweaks.
+        default_precision (str, optional): Default precision to use. Defaults to 'ap_fixed<16,6>'.
+        default_reuse_factor (int, optional): Default reuse factor. Defaults to 1.
+
+    Raises:
+        Exception: If Keras model has layers not supported by hls4ml.
+
+    Returns:
+        [dict]: The created config.
+    """
+    if granularity.lower() not in ['model', 'type', 'name']:
+        raise Exception('Invalid configuration granularity specified, expected "model", "type" or "name" got "{}"'.format(granularity))
 
     #This is a list of dictionaries to hold all the layer info we need to generate HLS
     layer_list = []
@@ -64,17 +100,17 @@ def config_from_keras_model(model, granularity='model', default_precision='ap_fi
     else:
         model_arch = json.loads(model.to_json())
 
-    #Define supported laers
-    core_layers = ['InputLayer', 'Dropout', 'Flatten', 'Reshape']
+    #Define supported layers
+    core_layers = ['InputLayer', 'Dropout', 'Flatten', 'Reshape', 'Permute']
     dense_layers = ['Dense', 'BinaryDense', 'TernaryDense']
     conv_layers = ['Conv1D', 'Conv2D', 'BinaryConv2D']
-    pooling_layers = ['MaxPooling1D', 'MaxPooling2D', 'AveragePooling1D', 'AveragePooling2D']
+    pooling_layers = ['MaxPooling1D', 'MaxPooling2D', 'GlobalMaxPooling1D', 'GlobalMaxPooling2D', 'AveragePooling1D', 'AveragePooling2D', 'GlobalAveragePooling1D', 'GlobalAveragePooling2D']
     norm_layers = ['BatchNormalization']
     activation_layers = ['Activation', 'LeakyReLU', 'ThresholdedReLU', 'ELU', 'PReLU', 'Softmax', 'ReLU']
-    merge_layers = ['Add', 'Subtract', 'Multiply', 'Average', 'Maximum', 'Minimum', 'Concatenate']
-    qkeras_layers = ['QDense', 'QActivation', 'QConv1D', 'QConv2D']
-    #Define layers to skip for conversion to HLS
-    skip_layers = ['Dropout', 'Flatten']
+    merge_layers = ['Add', 'Subtract', 'Multiply', 'Average', 'Maximum', 'Minimum', 'Concatenate', 'Dot']
+    qkeras_layers = ['QDense', 'QActivation', 'QConv1D', 'QConv2D', 'QBatchNormalization', 'QConv2DBatchnorm']
+    #Define layers to skip because they're not configurable or not converted to HLS
+    skip_layers = ['Dropout', 'Flatten', 'Reshape', 'Permute']
     #All supported layers
     supported_layers = core_layers + dense_layers + conv_layers + pooling_layers + norm_layers + activation_layers + merge_layers + qkeras_layers + skip_layers
 
@@ -185,17 +221,16 @@ def config_from_keras_model(model, granularity='model', default_precision='ap_fi
 
     config = {}
 
-    if granularity.lower() == 'model':
-        model_config = {}
-        model_config['Precision'] = default_precision
-        model_config['ReuseFactor'] = default_reuse_factor
-        model_config['Strategy'] = 'Latency'
-        #model_config['Compression'] = False
-        #model_config['Trace'] = False
-        
-        config['Model'] = model_config
+    model_config = {}
+    model_config['Precision'] = default_precision
+    model_config['ReuseFactor'] = default_reuse_factor
+    model_config['Strategy'] = 'Latency'
+    #model_config['Compression'] = False
+    #model_config['Trace'] = False
+
+    config['Model'] = model_config
     
-    elif granularity.lower() == 'type':
+    if granularity.lower() == 'type':
         type_config = {}
         for layer in layer_list:
             if layer['class_name'] in type_config:
@@ -213,7 +248,98 @@ def config_from_keras_model(model, granularity='model', default_precision='ap_fi
         
         config['LayerName'] = name_config
 
-    else:
-        raise Exception('Invalid configuration granularity specified, expected "model", "type" or "name" got "{}"'.format(granularity))
-
     return config
+
+
+def config_from_pytorch_model(model, granularity='model', default_precision='ap_fixed<16,6>', default_reuse_factor=1):
+    """Generate configuration dictionary from a Pytorch model.
+    
+    Parameters
+    ----------
+    model : Pytorch model object.
+        Model to be converted to hls model object.
+    granularity : string, optional
+        How granular you want the configuration to be.
+    default_precision : string, optional
+        Defines the precsion of your inputs, outputs, weights and biases.
+        It is denoted by ap_fixed<X,Y>, where Y is the number of bits representing 
+        the signed number above the binary point (i.e. the integer part),
+        and X is the total number of bits. Additionally, integers in fixed precision 
+        data type (ap_int<N>, where N is a bit-size from 1 to 1024) can also be used. 
+    default_reuse_factor : int, optional
+        Reuse factor for hls model
+        
+    Returns
+    -------
+    config : dict
+        configuration dictionary to be used in Pytorch converter.
+        
+    See Also
+    --------
+    hls4ml.config_from_keras_model, hls4ml.convert_from_onnx_model
+    
+    Examples
+    --------
+    >>> import hls4ml
+    >>> config = hls4ml.utils.config_from_keras_model(model, granularity='model')
+    >>> hls_model = hls4ml.converters.convert_from_keras_model(model, hls_config=config)
+
+    """
+    
+    config = {}
+
+    model_config = {}
+    model_config['Precision'] = default_precision
+    model_config['ReuseFactor'] = default_reuse_factor
+    model_config['Strategy'] = 'Latency'
+
+    config['Model'] = model_config
+    
+    return config
+
+
+def config_from_onnx_model(model, granularity='model', default_precision='ap_fixed<16,6>', default_reuse_factor=1):
+    """Generate configuration dictionary from an ONNX model.
+    
+    Parameters
+    ----------
+    model : ONNX model object.
+        Model to be converted to hls model object.
+    granularity : string, optional
+        How granular you want the configuration to be.
+    default_precision : string, optional
+        Defines the precsion of your inputs, outputs, weights and biases.
+        It is denoted by ap_fixed<X,Y>, where Y is the number of bits representing 
+        the signed number above the binary point (i.e. the integer part),
+        and X is the total number of bits. Additionally, integers in fixed precision 
+        data type (ap_int<N>, where N is a bit-size from 1 to 1024) can also be used. 
+    default_reuse_factor : int, optional
+        Reuse factor for hls model
+        
+    Returns
+    -------
+    config : dict
+        configuration dictionary to be used in ONNX converter.
+        
+    See Also
+    --------
+    hls4ml.config_from_keras_model, hls4ml.convert_from_pytorch_model
+    
+    Examples
+    --------
+    >>> import hls4ml
+    >>> config = hls4ml.utils.config_from_keras_model(model, granularity='model')
+    >>> hls_model = hls4ml.converters.convert_from_keras_model(model, hls_config=config)
+
+    """
+    
+    config = {}
+
+    model_config = {}
+    model_config['Precision'] = default_precision
+    model_config['ReuseFactor'] = default_reuse_factor
+    model_config['Strategy'] = 'Latency'
+
+    config['Model'] = model_config
+    
+    return config
\ No newline at end of file
