@@ -9,9 +9,49 @@ import glob
 from collections import OrderedDict
 
 from hls4ml.writer.writers import Writer
+from hls4ml.model.hls_layers import XnorPrecisionType
+
+config_filename = 'hls4ml_config.yml'
 
 class VivadoWriter(Writer):
 
+    def type_definition_cpp(self, model, atype):
+        type_class = atype.__class__.__name__
+        if type_class == 'HLSType':
+            return 'typedef {precision} {name};\n'.format(name=atype.name, precision=atype.precision)
+        elif type_class == 'CompressedType':
+            cpp_fmt = ('typedef struct {name} {{ '
+               '{index} row_index;'
+               '{index} col_index;'
+               '{precision} weight; }} {name};\n')
+            return cpp_fmt.format(name=atype.name, index=atype.index_precision, precision=atype.precision)
+        elif type_class == 'PackedType':
+            n_elem_expr = '/' if atype.unpack else '*'
+            return 'typedef nnet::array<{precision}, {n_elem}> {name};\n'.format(name=atype.name, precision=atype.precision, n_elem=str(atype.n_elem) + n_elem_expr + str(atype.n_pack))
+        elif type_class == 'ExponentType':
+            cpp_fmt = ('typedef struct {name} {{ '
+                       '{sign} sign; '
+                       '{precision} weight; }} {name};\n')
+            return cpp_fmt.format(name=atype.name, precision=atype.precision, sign=str(XnorPrecisionType()))
+        else:
+            raise Exception('Unknown data type class "{}"'.format(type_class))
+
+    def variable_definition_cpp(self, model, var, name_suffix='', as_reference=False):
+        var_class = var.__class__.__name__
+        if var_class == 'ArrayVariable':
+            return '{type} {name}{suffix}[{shape}]'.format(type=var.type.name, name=var.cppname, suffix=name_suffix, shape=var.size_cpp())
+        elif var_class == 'StreamVariable':
+            if as_reference: # Function parameter
+                return 'hls::stream<{type}> &{name}{suffix}'.format(type=var.type.name, name=var.cppname, suffix=name_suffix)
+            else: # Declaration
+                return 'hls::stream<{type}> {name}{suffix}("{name}")'.format(type=var.type.name, name=var.cppname, suffix=name_suffix)
+        elif var_class == 'WeightVariable':
+            return '{type} {name}{suffix}[{size}]'.format(type=var.type.name, name=var.cppname, suffix=name_suffix, size=var.data_length)
+        elif var_class == 'InplaceVariable':
+            return None
+        else:
+            raise Exception('Unknown variable class "{}"'.format(var_class))
+
     def print_array_to_cpp(self, var, odir, write_txt_file=True):
         #######################################
         ## Print weight array to C++
@@ -90,7 +130,7 @@ class VivadoWriter(Writer):
             return template.format(mode=mode.upper(), name=variable.name, type=typ, factor=factor, dim=0)
 
         elif mode == 'stream':
-            return '#pragma HLS STREAM variable={name} depth={depth} dim={dim}'.format(name=variable.name, depth=depth, dim=0)
+            return '#pragma HLS STREAM variable={name} depth={depth}'.format(name=variable.name, depth=depth)
 
     def write_project_cpp(self, model):
         ###################
@@ -103,6 +143,7 @@ class VivadoWriter(Writer):
 
         model_inputs = model.get_input_variables()
         model_outputs = model.get_output_variables()
+        model_brams   = model.get_bram_variables()
 
         indent = '    '
 
@@ -111,14 +152,17 @@ class VivadoWriter(Writer):
             if 'myproject' in line:
                 newline = line.replace('myproject', model.config.get_project_name())
             elif '//hls-fpga-machine-learning insert header' in line:
-                inputs_str = ', '.join([i.definition_cpp() for i in model_inputs])
-                outputs_str = ', '.join([o.definition_cpp() for o in model_outputs])
+                inputs_str = ', '.join([self.variable_definition_cpp(model, i, as_reference=True) for i in model_inputs])
+                outputs_str = ', '.join([self.variable_definition_cpp(model, o, as_reference=True) for o in model_outputs])
+                brams_str  = ', \n'.join([indent + self.variable_definition_cpp(model, b, as_reference=False) for b in model_brams])
                 insize_str = ', '.join(['unsigned short &const_size_in_{}'.format(i) for i in range(1, len(model_inputs) + 1)])
                 outsize_str = ', '.join(['unsigned short &const_size_out_{}'.format(i) for i in range(1, len(model_outputs) + 1)])
 
                 newline = ''
                 newline += indent + inputs_str + ',\n'
                 newline += indent + outputs_str + ',\n'
+                if len(model_brams) > 0: 
+                    newline += brams_str + ',\n'
                 newline += indent + insize_str + ',\n'
                 newline += indent + outsize_str + '\n'
 
@@ -128,6 +172,8 @@ class VivadoWriter(Writer):
                     for w in layer.get_weights():
                         if w.__class__.__name__ == 'CompressedWeightVariable':
                             newline += indent + '    nnet::load_compressed_weights_from_txt<{}, {}>({}, "{}.txt");\n'.format(w.type.name, w.nonzeros, w.name, w.name)
+                        elif w.__class__.__name__ == 'ExponentWeightVariable':
+                            newline += indent + '    nnet::load_exponent_weights_from_txt<{}, {}>({}, "{}.txt");\n'.format(w.type.name, w.data_length, w.name, w.name)
                         else:
                             newline += indent + '    nnet::load_weights_from_txt<{}, {}>({}, "{}.txt");\n'.format(w.type.name, w.data_length, w.name, w.name)
 
@@ -136,19 +182,23 @@ class VivadoWriter(Writer):
                 newline = line
                 all_inputs = [i.cppname for i in model_inputs]
                 all_outputs = [o.cppname for o in model_outputs]
+                all_brams = [b.cppname for b in model_brams] 
+                io_type = model.config.get_config_value("IOType")
 
-                if model.config.get_config_value("IOType") == "io_parallel":
+                if io_type == 'io_parallel':
                     for i in model_inputs: newline += indent + self._make_array_pragma(i) + '\n'
                     for o in model_outputs: newline += indent + self._make_array_pragma(o) + '\n'
                     # TODO discussed adding a handle for setting the interface mode for individual input and output arrays (16.03.2020)
                     # Probably the handle doesn't need to be exposed to the user but should be just set in hls_model.py
                     newline += indent + '#pragma HLS INTERFACE ap_vld port={},{} \n'.format(','.join(all_inputs), ','.join(all_outputs))
-                    if model.config.model_strategy == 'Resource':
+                    if model.config.model_strategy.lower() == 'resource':
                         newline += indent + '#pragma HLS DATAFLOW \n'
                     else:
                         newline += indent + '#pragma HLS PIPELINE \n'
-                if model.config.get_config_value("IOType") == "io_serial":
+                if io_type == 'io_serial' or io_type == 'io_stream':
                     newline += indent + '#pragma HLS INTERFACE axis port={},{} \n'.format(','.join(all_inputs), ','.join(all_outputs))
+                    if all_brams: # No BRAM ports
+                        newline += indent + '#pragma HLS INTERFACE bram port={} \n'.format(','.join(all_brams))
                     newline += indent + '#pragma HLS DATAFLOW \n'
 
                 inval_str = '\n    '.join(['const_size_in_{} = {};'.format(i, inp.size_cpp()) for i, inp in enumerate(model_inputs, 1)])
@@ -165,16 +215,20 @@ class VivadoWriter(Writer):
                     vars = layer.get_variables()
                     for var in vars:
                         if var not in inputs and var not in outputs:
-                            def_cpp = var.definition_cpp()
+                            def_cpp = self.variable_definition_cpp(model, var)
                             if def_cpp is not None:
                                 newline += '    ' + def_cpp + ';\n'
                                 if var.pragma:
                                     newline += '    ' + self._make_array_pragma(var) + '\n'
                     func = layer.function_cpp()
                     if func:
-                        for line in func:
-                            newline += '    ' + line + '\n'
-                        if model.config.trace_output and model.config.get_layer_config_value(layer, 'Trace', False):
+                        if len(func) == 1:
+                            newline += '    ' + func[0] + ' // ' + layer.name + '\n'
+                        else:
+                            newline += '// ' + layer.name + '\n'
+                            for line in func:
+                                newline += '    ' + line + '\n'
+                        if model.config.trace_output and layer.get_attr('Trace', False):
                             newline += '#ifndef __SYNTHESIS__\n'
                             for var in vars:
                                 newline += '    nnet::save_layer_output<{}>({}, "{}", {});\n'.format(var.type.name, var.name, layer.name, var.size_cpp())
@@ -201,6 +255,7 @@ class VivadoWriter(Writer):
 
         model_inputs = model.get_input_variables()
         model_outputs = model.get_output_variables()
+        model_brams   = model.get_bram_variables()
 
         indent = '    '
 
@@ -211,14 +266,17 @@ class VivadoWriter(Writer):
             elif 'void myproject(' in line:
                 newline = 'void {}(\n'.format(model.config.get_project_name())
             elif '//hls-fpga-machine-learning insert header' in line:
-                inputs_str = ', '.join([i.definition_cpp() for i in model_inputs])
-                outputs_str = ', '.join([o.definition_cpp() for o in model_outputs])
+                inputs_str = ', '.join([self.variable_definition_cpp(model, i, as_reference=True) for i in model_inputs])
+                outputs_str = ', '.join([self.variable_definition_cpp(model, o, as_reference=True) for o in model_outputs])
+                brams_str  = ', \n'.join([indent + self.variable_definition_cpp(model, b, as_reference=False) for b in model_brams])
                 insize_str = ', '.join(['unsigned short &const_size_in_{}'.format(i) for i in range(1, len(model_inputs) + 1)])
                 outsize_str = ', '.join(['unsigned short &const_size_out_{}'.format(o) for o in range(1, len(model_outputs) + 1)])
 
                 newline = ''
                 newline += indent + inputs_str + ',\n'
                 newline += indent + outputs_str + ',\n'
+                if len(model_brams) > 0: 
+                    newline += brams_str + ',\n'
                 newline += indent + insize_str + ',\n'
                 newline += indent + outsize_str + '\n'
             else:
@@ -248,8 +306,7 @@ class VivadoWriter(Writer):
                     layer_precision = layer.get_layer_precision()
                     all_precision.update(layer_precision)
                 for used_type in all_precision.values():
-                    newline += used_type.definition_cpp()
-
+                    newline += self.type_definition_cpp(model, used_type)
             else:
                 newline = line
             fout.write(newline)
@@ -272,13 +329,17 @@ class VivadoWriter(Writer):
                 newline = line
                 for layer in model.get_layers():
                     for w in layer.get_weights():
-                        newline += '#include "weights/{}.h"\n'.format(w.name)
+                        if w.name not in model.bram_vars.keys():
+                            newline += '#include "weights/{}.h"\n'.format(w.name)
+                        else:
+                            newline += '// #include "weights/{}.h"\n'.format(w.name)
 
             elif "//hls-fpga-machine-learning insert layer-config" in line:
                 newline = line
                 for layer in model.get_layers():
                     config = layer.config_cpp()
                     if config:
+                        newline += '// ' + layer.name + '\n'
                         newline += config + '\n'
             else:
                 newline = line
@@ -351,28 +412,28 @@ class VivadoWriter(Writer):
             #Insert numbers
             if 'myproject' in line:
                 newline = line.replace('myproject', model.config.get_project_name())
+            elif '//hls-fpga-machine-learning insert bram' in line:
+                newline = line
+                for bram in model.get_bram_variables():
+                    newline += '#include \"firmware/weights/{}.h\"\n'.format(bram.cppname)
             elif '//hls-fpga-machine-learning insert data' in line:
                 newline = line
-                newline += '      std::vector<float>::const_iterator in_begin = in.cbegin();\n'
-                newline += '      std::vector<float>::const_iterator in_end;\n'
+                offset = 0
+                # for bram in model.get_bram_variables():
+                #     newline += bram.definition_cpp()+';\n'
                 for inp in model.get_input_variables():
-                    newline += '      ' + inp.definition_cpp() + ';\n'
-                    newline += '      in_end = in_begin + ({});\n'.format(inp.size_cpp())
-                    newline += '      std::copy(in_begin, in_end, {});\n'.format(inp.cppname)
-                    newline += '      in_begin = in_end;\n'
+                    newline += '      ' + self.variable_definition_cpp(model, inp) + ';\n'
+                    newline += '      nnet::copy_data<float, {}, {}, {}>(in, {});\n'.format(inp.type.name, offset, inp.size_cpp(), inp.cppname)
+                    offset += inp.size()
                 for out in model.get_output_variables():
-                    # brace-init zeros the array out because we use std=c++0x
-                    newline += '      ' + out.definition_cpp() + '{};\n'
-                    # but we can still explicitly zero out if you want
-                    newline += '      std::fill_n({}, {}, 0.);\n'.format(out.cppname, out.size())
+                    newline += '      ' + self.variable_definition_cpp(model, out) + ';\n'
             elif '//hls-fpga-machine-learning insert zero' in line:
                 newline = line
                 for inp in model.get_input_variables():
-                    newline += '    ' + inp.definition_cpp() + ';\n'
-                    newline += '    std::fill_n({}, {}, 0.);\n'.format(inp.cppname, inp.size_cpp())
+                    newline += '    ' + self.variable_definition_cpp(model, inp) + ';\n'
+                    newline += '    nnet::fill_zero<{}, {}>({});\n'.format(inp.type.name, inp.size_cpp(), inp.cppname)
                 for out in model.get_output_variables():
-                    newline += '    ' + out.definition_cpp() + '{};\n'
-                    newline += '      std::fill_n({}, {}, 0.);\n'.format(out.cppname, out.size())
+                    newline += '    ' + self.variable_definition_cpp(model, out) + ';\n'
             elif '//hls-fpga-machine-learning insert top-level-function' in line:
                 newline = line
 
@@ -383,7 +444,13 @@ class VivadoWriter(Writer):
 
                 input_vars = ','.join([i.cppname for i in model.get_input_variables()])
                 output_vars = ','.join([o.cppname for o in model.get_output_variables()])
-                top_level = indent + '{}({},{},{},{});\n'.format(model.config.get_project_name(), input_vars, output_vars, input_size_vars, output_size_vars)
+                bram_vars   =','.join([b.cppname for b in model.get_bram_variables()]) 
+
+                # Concatenate the input, output, and bram variables. Filter out empty/null values
+                all_vars = ','.join(filter(None, [input_vars, output_vars, bram_vars]))
+
+                top_level = indent + '{}({},{},{});\n'.format(model.config.get_project_name(), all_vars, input_size_vars, output_size_vars)
+                
                 newline += top_level
             elif '//hls-fpga-machine-learning insert predictions' in line:
                 newline = line
@@ -395,17 +462,11 @@ class VivadoWriter(Writer):
             elif '//hls-fpga-machine-learning insert tb-output' in line:
                 newline = line
                 for out in model.get_output_variables():
-                    newline += indent + 'for(int i = 0; i < {}; i++) {{\n'.format(out.size_cpp())
-                    newline += indent + '  fout << {}[i] << " ";\n'.format(out.cppname)
-                    newline += indent + '}\n'
-                    newline += indent + 'fout << std::endl;\n'
+                    newline += indent + 'nnet::print_result<{}, {}>({}, fout);\n'.format(out.type.name, out.size_cpp(), out.cppname) #TODO enable this
             elif '//hls-fpga-machine-learning insert output' in line or '//hls-fpga-machine-learning insert quantized' in line:
                 newline = line
                 for out in model.get_output_variables():
-                    newline += indent + 'for(int i = 0; i < {}; i++) {{\n'.format(out.size_cpp())
-                    newline += indent + '  std::cout << {}[i] << " ";\n'.format(out.cppname)
-                    newline += indent + '}\n'
-                    newline += indent + 'std::cout << std::endl;\n'
+                    newline += indent + 'nnet::print_result<{}, {}>({}, std::cout, true);\n'.format(out.type.name, out.size_cpp(), out.cppname)
             else:
                 newline = line
             fout.write(newline)
@@ -423,6 +484,7 @@ class VivadoWriter(Writer):
 
         model_inputs = model.get_input_variables()
         model_outputs = model.get_output_variables()
+        model_brams   = model.get_bram_variables()
 
         indent = '    '
 
@@ -432,6 +494,10 @@ class VivadoWriter(Writer):
                 newline = line.replace('MYPROJECT', format(model.config.get_project_name().upper()))
             elif 'myproject' in line:
                 newline = line.replace('myproject', format(model.config.get_project_name()))
+            elif '//hls-fpga-machine-learning insert bram' in line:
+                newline = line
+                for bram in model_brams:
+                    newline += '#include \"firmware/weights/{}.h\"\n'.format(bram.cppname)
             elif '//hls-fpga-machine-learning insert header' in line:
                 dtype = line.split('#', 1)[1].strip()
                 inputs_str = ', '.join(['{type} {name}[{shape}]'.format(type=dtype, name=i.cppname, shape=i.size_cpp()) for i in model_inputs])
@@ -448,20 +514,25 @@ class VivadoWriter(Writer):
                 dtype = line.split('#', 1)[1].strip()
                 newline = ''
                 for i in model_inputs:
-                    newline += indent + '{type} {name}_ap[{shape}];\n'.format(type=i.type.name, name=i.cppname, shape=i.size_cpp())
+                    newline += indent + '{var};\n'.format(var=self.variable_definition_cpp(model, i, name_suffix='_ap'))
                     newline += indent + 'nnet::convert_data<{}, {}, {}>({}, {}_ap);\n'.format(dtype, i.type.name, i.size_cpp(), i.cppname, i.cppname)
                 newline += '\n'
                 
                 for o in model_outputs:
-                    newline += indent + '{type} {name}_ap[{shape}];\n'.format(type=o.type.name, name=o.cppname, shape=o.size_cpp())
+                    newline += indent + '{var};\n'.format(var=self.variable_definition_cpp(model, o, name_suffix='_ap'))
                 
                 newline += '\n'
 
                 input_size_vars = ','.join(['const_size_in_{}'.format(i) for i in range(1, len(model.get_input_variables()) + 1)])
                 output_size_vars = ','.join(['const_size_out_{}'.format(o) for o in range(1, len(model.get_output_variables()) + 1)])
                 input_vars = ','.join([i.cppname + '_ap' for i in model.get_input_variables()])
+                bram_vars   =','.join([b.cppname for b in model.get_bram_variables()]) 
                 output_vars = ','.join([o.cppname + '_ap' for o in model.get_output_variables()])
-                top_level = indent + '{}({}, {}, {}, {});\n'.format(model.config.get_project_name(), input_vars, output_vars, input_size_vars, output_size_vars)
+                
+                # Concatenate the input, output, and bram variables. Filter out empty/null values
+                all_vars = ','.join(filter(None, [input_vars, output_vars, bram_vars]))
+
+                top_level = indent + '{}({},{},{});\n'.format(model.config.get_project_name(), all_vars, input_size_vars, output_size_vars)
                 newline += top_level
 
                 newline += '\n'
@@ -471,7 +542,7 @@ class VivadoWriter(Writer):
             elif '//hls-fpga-machine-learning insert trace_outputs' in line:
                 newline = ''
                 for layer in model.get_layers():
-                    if layer.function_cpp() and model.config.trace_output and model.config.get_layer_config_value(layer, 'Trace', False):
+                    if layer.function_cpp() and model.config.trace_output and layer.get_attr('Trace', False):
                             vars = layer.get_variables()
                             for var in vars:
                                 newline += indent + 'nnet::trace_outputs->insert(std::pair<std::string, void *>("{}", (void *) malloc({} * element_size)));\n'.format(layer.name, var.size_cpp())
@@ -569,6 +640,25 @@ class VivadoWriter(Writer):
 
         copytree(srcpath, dstpath)
 
+    def write_yml(self, model):
+        ###################
+        # YAML config file
+        ###################
+
+        def keras_model_representer(dumper, keras_model):
+            model_path = model.config.get_output_dir() + '/keras_model.h5'
+            keras_model.save(model_path)
+            return dumper.represent_scalar(u'!keras_model', model_path)
+
+        try:
+            from tensorflow.keras import Model as KerasModel
+            yaml.add_multi_representer(KerasModel, keras_model_representer)
+        except:
+            pass
+
+        with open(model.config.get_output_dir() + '/' + config_filename, 'w') as file:
+            yaml.dump(model.config.config, file)
+
     def write_tar(self, model):
         ###################
         # Tarball output
@@ -589,5 +679,6 @@ class VivadoWriter(Writer):
         self.write_bridge(model)
         self.write_build_script(model)
         self.write_nnet_utils(model)
+        self.write_yml(model)
         self.write_tar(model)
         print('Done')
