@@ -3,7 +3,6 @@
 models_var=HLS4ML_KERAS_MODELS
 models_file=
 exec=echo
-py=
 dir=
 
 function print_usage {
@@ -18,22 +17,18 @@ function print_usage {
    echo "      environment variable."
    echo "   -x"
    echo "      Execute the commands instead of just printing them."
-   echo "   -p 2|3"
-   echo "      Python version passed to keras-to-hls script (2 or 3)."
    echo "   -d DIR"
    echo "      Output directory passed to keras-to-hls script."
    echo "   -h"
    echo "      Prints this help message."
 }
 
-while getopts ":f:xp:d:h" opt; do
+while getopts ":f:xd:h" opt; do
    case "$opt" in
    f) models_file=$OPTARG
       ;;
    x) exec=eval
       ;;
-   p) py="-p $OPTARG"
-      ;;
    d) dir="-d $OPTARG"
       ;;
    h)
@@ -60,22 +55,24 @@ fi
 
 for line in "${model_line[@]}"
 do
-   params=("" "" "" "" "" "" "")
+   params=("" "" "" "" "" "" "" "" "" "")
    if [[ ${line} = *[![:space:]]* ]] && ! [[ "${line}" = \#* ]] ; then
       IFS=" " read -ra model_def <<< "${line}"
       for (( i=1; i<"${#model_def[@]}"; i++ ));
       do
          if [[ "${model_def[$i]}" == x:* ]] ; then params[0]="-x ${model_def[$i]:2} "; fi
-         if [[ "${model_def[$i]}" == c:* ]] ; then params[1]="-c ${model_def[$i]:2} "; fi
-         if [[ "${model_def[$i]}" == io:s ]] ; then params[2]="-s "; fi
-         if [[ "${model_def[$i]}" == r:* ]] ; then params[3]="-r ${model_def[$i]:2} "; fi
-         if [[ "${model_def[$i]}" == s:* ]] ; then params[4]="-g ${model_def[$i]:2} "; fi
-         if [[ "${model_def[$i]}" == i:* ]] ; then params[5]="-t ${model_def[$i]:2} "; fi
+         if [[ "${model_def[$i]}" == b:* ]] ; then params[1]="-b ${model_def[$i]:2} "; fi
+         if [[ "${model_def[$i]}" == B:* ]] ; then params[2]="-B ${model_def[$i]:2} "; fi
+         if [[ "${model_def[$i]}" == c:* ]] ; then params[3]="-c ${model_def[$i]:2} "; fi
+         if [[ "${model_def[$i]}" == io:s ]] ; then params[4]="-s "; fi
+         if [[ "${model_def[$i]}" == r:* ]] ; then params[5]="-r ${model_def[$i]:2} "; fi
+         if [[ "${model_def[$i]}" == s:* ]] ; then params[6]="-g ${model_def[$i]:2} "; fi
+         if [[ "${model_def[$i]}" == i:* ]] ; then params[7]="-t ${model_def[$i]:2} "; fi
+         if [[ "${model_def[$i]}" == y:* ]] ; then params[8]="-y ${model_def[$i]:2} "; fi
       done
-      params[6]=${model_def[0]}
-
-      cmd="./keras-to-hls.sh ${py} ${dir} ${params[0]}${params[1]}${params[2]}${params[3]}${params[4]}${params[5]}${params[6]}"
-
+      params[9]=${model_def[0]}
+      cmd="./keras-to-hls.sh ${dir} ${params[0]}${params[1]}${params[2]}${params[3]}${params[4]}${params[5]}${params[6]}${params[7]}${params[8]}${params[9]}"
+      echo "$cmd"
       ${exec} "${cmd}"
    fi
 done
