@@ -1,12 +1,15 @@
 #!/bin/bash
 
 pycmd=python
-xilinxpart="xc7vx690tffg1927-2"
+xilinxpart="xcvu9p-flgb2104-2-e"
+board="None"
+backend="Vivado"
 clock=5
 io=io_parallel
 rf=1
 strategy="Latency"
 type="ap_fixed<16,6>"
+yml=""
 basedir=vivado_prj
 
 sanitizer="[^A-Za-z0-9._]"
@@ -20,15 +23,16 @@ function print_usage {
    echo "Multiple models can be specified."
    echo ""
    echo "Options are:"
-   echo "   -p 2|3"
-   echo "      Python version to use (2 or 3). If not specified uses default"
-   echo "      'python' interpreter."
-   echo "   -x DEVICE"
-   echo "      Xilinx device part number. Defaults to 'xc7vx690tffg1927-2'."
+   echo "   -x XILINXPART"
+   echo "      Xilinx device part number. Defaults to 'xcvu9p-flgb2104-2-e'."
+   echo "   -b BOARD"
+   echo "      Board used. Defaults to 'pynq-z2'."
+   echo "   -B BACKEND"
+   echo "      Backend to use for the generation of the code. Defaults to 'Vivado'."
    echo "   -c CLOCK"
    echo "      Clock period to use. Defaults to 5."
    echo "   -s"
-   echo "      Use serial I/O. If not specified uses parallel I/O."
+   echo "      Use streaming I/O. If not specified uses parallel I/O."
    echo "   -r FACTOR"
    echo "      Reuse factor. Defaults to 1."
    echo "   -g STRATEGY"
@@ -37,19 +41,23 @@ function print_usage {
    echo "      Default precision. Defaults to 'ap_fixed<16,6>'."
    echo "   -d DIR"
    echo "      Output directory."
+   echo "   -y FILE"
+   echo "      YAML config file to take HLS config from. If specified, -r, -g and -t are ignored."
    echo "   -h"
    echo "      Prints this help message."
 }
 
-while getopts ":p:x:c:sr:g:t:d:h" opt; do
+while getopts ":x:b:B:c:sr:g:t:d:y:h" opt; do
    case "$opt" in
-   p) pycmd=${pycmd}$OPTARG
-      ;;
    x) xilinxpart=$OPTARG
       ;;
+   b) board=$OPTARG
+      ;;
+   B) backend=$OPTARG
+      ;;
    c) clock=$OPTARG
       ;;
-   s) io=io_serial
+   s) io=io_stream
       ;;
    r) rf=$OPTARG
       ;;
@@ -59,6 +67,8 @@ while getopts ":p:x:c:sr:g:t:d:h" opt; do
       ;;
    d) basedir=$OPTARG
       ;;
+   y) yml=$OPTARG
+      ;;
    h)
       print_usage
       exit
@@ -92,23 +102,38 @@ do
 
    echo "Creating config file for model '${model}'"
    base=`echo "${h5}" | sed -e 's/\(_weights\)*$//g'`
-   file="${basedir}/${base}-${pycmd}.yml"
+   file="${basedir}/${base}.yml"
+   prjdir="${basedir}/${base}-backend${backend}-board${board//${sanitizer}/_}-${xilinxpart//${sanitizer}/_}-c${clock}-${io}-rf${rf}-${type//${sanitizer}/_}-${strategy}"
+
+   hlscfg=""
+   if [ ! -z "${yml}" ]; then
+      hlscfg=`sed -ne '/HLSConfig/,$p' ../example-models/config-files/${yml}`
+   fi
 
    echo "KerasJson: ../example-models/keras/${name}.json" > ${file}
    echo "KerasH5:   ../example-models/keras/${h5}.h5" >> ${file}
-   echo "OutputDir: ${basedir}/${base}-${pycmd}-${xilinxpart//${sanitizer}/_}-c${clock}-${io}-rf${rf}-${type//${sanitizer}/_}-${strategy}" >> ${file}
+   echo "OutputDir: ${prjdir}" >> ${file}
    echo "ProjectName: myproject" >> ${file}
    echo "XilinxPart: ${xilinxpart}" >> ${file}
+   echo "Board: ${board}" >> ${file}
+   echo "Backend: ${backend}" >> ${file}
    echo "ClockPeriod: ${clock}" >> ${file}
    echo "" >> ${file}
    echo "IOType: ${io}" >> ${file}
-   echo "HLSConfig:" >> ${file}
-   echo "  Model:" >> ${file}
-   echo "    ReuseFactor: ${rf}" >> ${file}
-   echo "    Precision: ${type} " >> ${file}
-   echo "    Strategy: ${strategy} " >> ${file}
+   
+   if [ -z "${hlscfg}" ]
+   then
+      echo "HLSConfig:" >> ${file}
+      echo "  Model:" >> ${file}
+      echo "    ReuseFactor: ${rf}" >> ${file}
+      echo "    Precision: ${type} " >> ${file}
+      echo "    Strategy: ${strategy} " >> ${file}
+   else
+      echo "${hlscfg}" >> ${file}
+   fi
 
    ${pycmd} ../scripts/hls4ml convert -c ${file} || exit 1
    rm ${file}
+   rm -rf "${prjdir}"
    echo ""
 done
